getAccessTokenPKCE() {
  console.log('pkce flow');
  const state = generateRandomString();
  const codeVerifier = generateRandomString();
  const redirectUrl = getRedirectUrl(this.provider);
  return getPkceChallenge(codeVerifier)
    .then(codeChallenge => {
      console.log('pkce flow', codeChallenge);
      let url = this.authorizationEndpoint;
      url += `?response_type=${this.responseType}`;
      url += `&scope=${encodeURIComponent(this.scopes.join(' '))}`;
      url += `&client_id=${this.clientId}`;
      url += `&redirect_uri=${encodeURIComponent(redirectUrl)}`;
      url += `&state=${state}`;
      url += "&code_challenge="+encodeURIComponent(codeChallenge);
      url += "&code_challenge_method=S256";
      console.log(url);
      return browser.identity.launchWebAuthFlow({ interactive: true, url });
    })
    .then(url => {
      console.log('pkce flow response url', url);
      let params;
      const parsedUrl = new URL(url);
      if (parsedUrl.search) {
        // Try parse from query params first
        const searchParams = new URLSearchParams(parsedUrl.search);
        params = {
          code: searchParams.get('code'),
          state: searchParams.get('state')
        };
      } else {
        // parse params from hash fragment
        params = getParamsFromCallbackUrl(url);
      }
      if (params.state !== state) {
        throw new Error('Invalid state.');
      }
      const payload = new URLSearchParams();
      payload.set('grant_type', 'authorization_code');
      payload.set('code', params.code);
      payload.set('client_id', this.clientId);
      payload.set('redirect_uri', redirectUrl);
      payload.set('code_verifier', codeVerifier);
      return fetch(this.tokenEndpoint, {
        method: 'POST',
        headers: {
          'Content-Type': 'application/x-www-form-urlencoded'
        },
        body: payload
      });
    })
    .then(({ access_token, refresh_token }) => 
      this.storage.set(ACESS_TOKEN_KEY, access_token)
        .then(() => this.storage.set(REFRESH_TOKEN_KEY, refresh_token))
        .then(() => access_token))
    .catch(this.promptErrorHandler);
}



const sha256 = plain => {
  const encoder = new TextEncoder();
  const data = encoder.encode(plain);
  return window.crypto.subtle.digest('SHA-256', data);
};

const base64urlencode = str => {
  return btoa(String.fromCharCode.apply(null, new Uint8Array(str)))
      .replace(/\+/g, '-').replace(/\//g, '_').replace(/=+$/, '');
};

export default verifier => {
  return sha256(verifier).then(hashed => base64urlencode(hashed));
};


refreshToken(refreshToken) {
  const payload = new URLSearchParams();
  payload.set('client_id', this.clientId);
  payload.set('scope', encodeURIComponent(this.scopes.join(' ')));
  payload.set('refresh_token', refreshToken);
  payload.set('redirect_uri', encodeURIComponent(getRedirectUrl(this.provider)));
  payload.set('grant_type', 'refresh_token');
  return enhancedFetch(this.tokenEndpoint, {
    method: 'POST',
    headers: {
      'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
    },
    body: payload
  })
  .then(({ access_token, refresh_token }) => 
    this.storage.set(ACESS_TOKEN_KEY, access_token)
      .then(() => this.storage.set(REFRESH_TOKEN_KEY, refresh_token))
      .then(() => access_token))
      .catch(this.promptErrorHandler);
}
